# -- import packages: ---------------------------------------------------------
import pandas as pd
import pathlib
import wandb

# -- set type hints: ----------------------------------------------------------
from typing import Optional, Union

# -- function: ----------------------------------------------------------------
def download_run_history(
    run: wandb.apis.public.Run, fpath: Optional[Union[pathlib.Path, str]] = None
) -> pd.DataFrame:
    """Download the history of a run as a pandas DataFrame and save to .csv.

    Args:
        run: The run to download the history from.
        fpath: The path to save the history to.

    Returns:
        A pandas DataFrame containing the history of the run.
    """
    scan_history = run.scan_history()
    history = [item for item in scan_history]
    df = pd.DataFrame(history)
    if not fpath is None:
        df.to_csv(fpath)
    return df



# -- import packages: ---------------------------------------------------------
import ABCParse
import pandas as pd
import pathlib
import scdiffeq as sdq


# -- set typing: --------------------------------------------------------------
from typing import List


# -- operational object class: ------------------------------------------------
class VersionAccuracy(ABCParse.ABCParse):
    def __init__(
        self,
        version: sdq.io.Version,
        train_col: str = "unique_train.all_fates",
        test_col: str = "unique_test.all_fates",
        *args,
        **kwargs,
    ):
        self.__parse__(locals())

    @property
    def csv_paths(self) -> List[pathlib.Path]:
        return list(self._version._PATH.glob("fate_prediction_metrics/*/accuracy.csv"))

    def _read_frame(self, path: pathlib.Path):

        name = path.parent.name
        return (
            pd.read_csv(path, index_col=0)
            .loc[[self._train_col, self._test_col]]
            .rename({self._train_col: "train", self._test_col: "test"}, axis=0)
            .rename({"accuracy": name}, axis=1)
        )

    def _format_ckpt_name(self, name):
        return name.replace("=", "_").split(".ckpt")[0].replace("-", ".")

    @property
    def _saved_ckpt_fpaths(self) -> List[pathlib.Path]:
        if not hasattr(self, "_cached_saved_ckpt_fpaths"):
            raw_paths: List[pathlib.Path] = []
            if self._version and hasattr(self._version, 'ckpts') and self._version.ckpts:
                items_to_iterate = []
                if isinstance(self._version.ckpts, dict):
                    items_to_iterate = self._version.ckpts.values()
                elif isinstance(self._version.ckpts, list):
                    items_to_iterate = self._version.ckpts
                
                for ckpt_obj in items_to_iterate:
                    if hasattr(ckpt_obj, 'path') and isinstance(getattr(ckpt_obj, 'path'), pathlib.Path):
                        raw_paths.append(getattr(ckpt_obj, 'path'))
            self._cached_saved_ckpt_fpaths = raw_paths
        return self._cached_saved_ckpt_fpaths

    @property
    def _saved_ckpts(self):
        """Not all (e.g., `on_train_epoch_end`) ckpts get saved. Use this to filter accordingly"""
        if not hasattr(self, "_cached_saved_ckpts_names"):
            self._cached_saved_ckpts_names = [self._format_ckpt_name(p.name) for p in self._saved_ckpt_fpaths]
        return self._cached_saved_ckpts_names

    @property
    def _ckpt_name_to_path_mapping(self) -> dict[str, pathlib.Path]:
        if not hasattr(self, "_internal_ckpt_name_to_path_map_cache"):
            mapping = {}
            for p in self._saved_ckpt_fpaths:  # These are Path objects from _saved_ckpt_fpaths
                formatted_name = self._format_ckpt_name(p.name)
                path_to_store = p  # Default to the original path

                if p.name == "last.ckpt": # Only apply special logic for "last.ckpt"
                    if p.is_symlink():
                        try:
                            # Attempt to resolve the symlink strictly (target must exist)
                            resolved_path = p.resolve(strict=True)
                            path_to_store = resolved_path
                        except FileNotFoundError:
                            # Symlink is broken or target doesn't exist.
                            # Try to read what it points to and construct the intended path on the CURRENT machine.
                            try:
                                target_link_str = p.readlink() # Gets the path the symlink points to as a string
                                target_path_obj = pathlib.Path(target_link_str) # Convert to Path object

                                if target_path_obj.is_absolute():
                                    # If the link target was an absolute path (from another machine),
                                    # we only care about its filename and assume it should be
                                    # relative to the symlink's location on *this* machine.
                                    actual_filename = target_path_obj.name
                                    intended_path = (p.parent / actual_filename).resolve(strict=False)
                                else:
                                    # If the link target was relative, interpret it relative to the symlink's parent.
                                    intended_path = (p.parent / target_path_obj).resolve(strict=False)
                                
                                path_to_store = intended_path
                                # print(f"Warning: Symlink {p} is broken. Using re-interpreted target: {intended_path}")
                            except Exception as e_readlink_reconstruct:
                                # Failed to readlink or reconstruct path (e.g., permissions, or not a symlink after all)
                                # print(f"Warning: Could not reconstruct path for broken symlink {p}: {e_readlink_reconstruct}. Using original path {p}.")
                                path_to_store = p # Fallback to original symlink path
                        except Exception as e_resolve:
                            # Other errors during p.resolve(strict=True) (e.g. permission errors not FileNotFoundError)
                            # print(f"Warning: Error resolving symlink {p} strictly: {e_resolve}. Using original path {p}.")
                            path_to_store = p # Fallback to original symlink path
                    # else: # It's a regular file named "last.ckpt", not a symlink.
                        # print(f"Info: 'last.ckpt' at {p} is a regular file, not a symlink. Using its direct path.")
                        # path_to_store remains p, which is correct for a non-symlink.
                
                mapping[formatted_name] = path_to_store
            self._internal_ckpt_name_to_path_map_cache = mapping
        return self._internal_ckpt_name_to_path_map_cache

    def _get_ckpt_epoch_values(self, df_with_paths: pd.DataFrame) -> List[int]:
        epochs = []
        for original_idx_name, row in df_with_paths.iterrows():
            # original_idx_name is from df.index, e.g., "last", "epoch_123.v1"
            # Split off potential versioning like ".v1" to get the base epoch name part
            current_epoch_str_part = original_idx_name.split('.')[0]  # "epoch_123.v1" -> "epoch_123"; "last" -> "last"

            if current_epoch_str_part == "last":
                ckpt_path_obj = row["ckpt_path"]  # This is a pathlib.Path, potentially resolved
                actual_file_name = ckpt_path_obj.name  # e.g., "epoch_2499.ckpt" or "last.ckpt"
                
                # Use _format_ckpt_name to standardize the name before parsing epoch
                # e.g., "epoch_2499.ckpt" -> "epoch_2499"
                formatted_name_for_epoch_parsing = self._format_ckpt_name(actual_file_name)

                if formatted_name_for_epoch_parsing.startswith("epoch_"):
                    try:
                        # Extract epoch number, e.g., from "epoch_2499"
                        epoch_val = int(formatted_name_for_epoch_parsing.split("epoch_")[1])
                        epochs.append(epoch_val)
                    except (ValueError, IndexError):
                        # Fallback if parsing from resolved name (e.g. "epoch_XYZ") fails
                        epochs.append(2500) 
                else:
                    # Fallback if resolved name isn't "epoch_XXX" (e.g., if it's still "last" or something else)
                    epochs.append(2500)
            elif current_epoch_str_part.startswith("epoch_"):
                try:
                    # Extract epoch number from names like "epoch_123"
                    epoch_val = int(current_epoch_str_part.split("epoch_")[1])
                    epochs.append(epoch_val)
                except (ValueError, IndexError):
                    # Fallback for regular epoch parsing failure
                    # print(f"Warning: Could not parse epoch from '{current_epoch_str_part}'")
                    epochs.append(-1)  # Placeholder for unparsable epoch
            else:
                # Handle other non-"epoch_" or "last" names if they exist
                # print(f"Warning: Unrecognized epoch format: '{current_epoch_str_part}'")
                epochs.append(-1)  # Placeholder for unknown formats
        return epochs

    @property
    def df(self):
        if not hasattr(self, "_df"):
            df_parts = [self._read_frame(path) for path in self.csv_paths]
            
            if not df_parts:
                self._df = pd.DataFrame(columns=["train", "test", "ckpt_path", "epoch"])
                return self._df

            df = pd.concat(
                df_parts, axis=1
            ).T
            
            df = df.loc[df.index.isin(self._saved_ckpts)]
            
            df["ckpt_path"] = df.index.map(self._ckpt_name_to_path_mapping)
            
            df.dropna(subset=["ckpt_path"], inplace=True)
            
            df["epoch"] = self._get_ckpt_epoch_values(df)
            self._df = df.sort_values(["train", "epoch"], ascending=[False, False])
        return self._df

    @property
    def best_training_ckpt(self):
        return self.df.index[0]

    @property
    def best_test_from_train(self) -> pd.Series:
        return self.df.loc[self.best_training_ckpt]



# -- import packages: ---------------------------------------------------------
import anndata
import ABCParse
import anndata
import lightning
import scdiffeq as sdq
import uuid
import wandb

# -- import local dependencies: -----------------------------------------------
from ._gmm_sampling import GaussianMixtureModelSampling

# -- cls: ---------------------------------------------------------------------
class ExperimentRunner(ABCParse.ABCParse):

    def __init__(
        self,
        wandb_api_key: str,
        project_name: str,
        adata: anndata.AnnData,
        n_components: int = 50,
        n_gmm_components: int = 20,
        random_state: int = 42,
        ckpt_frequency: int = 1,
        save_last_ckpt: bool = True,
        keep_ckpts: int = 3,
        train_epochs: int = 2500,
        monitor: str = "epoch_validation_loss",
        pseudotime_key: str = "ct_pseudotime",
    ) -> None:

        self.__parse__(locals())

        self._run_id = str(uuid.uuid4()).split("-")[0]

        wandb.login(key=self._wandb_api_key)

    def _setup(self, N: int = 20_000) -> anndata.AnnData:
        self._gmm_sampling = GaussianMixtureModelSampling(
            adata=self._adata,
            n_components=self._n_components,
            n_gmm_components=self._n_gmm_components,
            random_state=self._random_state,
        )
        adata_sim = self._gmm_sampling(target_size=N)
        sdq.tl.bin_pseudotime(adata_sim, pseudotime_key = self._pseudotime_key, n_bins=self._n_bins)
        return adata_sim

    def forward(self, N: int = 20_000, seed: int = 0) -> None:
        """ """

        adata_sim = self._setup(N=N)

        wandb.init(project=self._project_name)
        wandb_logger = lightning.pytorch.loggers.WandbLogger(
            project=self._project_name
        )

        MODEL_PARAMS = {
            "adata": adata_sim,
            "use_key": "X",
            "time_key": "t",
            "mu_hidden": [512, 512],
            "sigma_hidden": [32, 32],
            "train_lr": 1e-4,
            "train_step_size": 1500,
            "monitor_hardware": True,
            "dt": self._dt,
            "batch_size": self._batch_size,
            "train_val_split": [0.9, 0.1],
            "seed": seed,
            "latent_dim": 50,
            "velocity_ratio_params": {
                "target": 2.5,
                "enforce": 100,
                "method": "square",
            },
            "potential_type": "fixed",
        }

        wandb_params = {k: v for k, v in MODEL_PARAMS.items() if k != "adata"}
        wandb_logger.experiment.config.update(wandb_params)
        wandb_logger.experiment.config.update(
            {
                "run_id": self._run_id,
                "N": N,
                "n_bins": self._n_bins,
                "distance_mean": self._gmm_sampling._distance_mean,
                "distance_std": self._gmm_sampling._distance_std,
            }
        )
        MODEL_PARAMS["logger"] = wandb_logger

        model = sdq.scDiffEq(**MODEL_PARAMS)

        model.fit(
            train_epochs=self._train_epochs,
            train_callbacks=[sdq.callbacks.StochasticWeightAveraging(swa_lrs=[1e-5])],
            ckpt_frequency=self._ckpt_frequency,
            save_last_ckpt=self._save_last_ckpt,
            keep_ckpts=self._keep_ckpts,
            monitor=self._monitor,
            mode="min",
        )

        wandb.finish()

    def __call__(
        self,
        N: int = 20_000,
        n_bins: int = 10,
        seed: int = 0,
        batch_size: int = 2048,
        dt: float = 0.1,
    ) -> None:
        """ """

        self.__update__(locals())

        self.forward(N=N, seed=seed)



from ._gmm_sampling import GaussianMixtureModelSampling
from ._map_time import map_time
from ._experiment_runner import ExperimentRunner

__all__ = ["GaussianMixtureModelSampling", "map_time", "ExperimentRunner"]


# -- import packages: ---------------------------------------------------------
import pandas as pd
import scdiffeq as sdq


# -- import local dependencies: -----------------------------------------------
from ._summarized_ckpt import SummarizedCkpt


# -- API-facing function: -----------------------------------------------------
def best_checkpoints(project: sdq.io.Project) -> pd.DataFrame:
    """Summarize the best checkpoint for each version of a project.

    Args:
        project (sdq.io.Project): An scdiffeq project instance.

    Returns:
        best_ckpts (pd.DataFrame): A DataFrame summarizing the best checkpoint for each version.
    """

    results = SummarizedCkpt(project=project)
    return results()



# -- import packages: ---------------------------------------------------------
import json
import logging
import pathlib
import pandas as pd
import wandb

# -- import local dependencies: -----------------------------------------------
from ._download_run_history import download_run_history

# -- set type hints: ----------------------------------------------------------
from typing import Any, Dict, List, Union

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)

# -- object cls: --------------------------------------------------------------
class Run:
    _DOWNLOAD_CALLED = False

    def __init__(
        self,
        run: wandb.apis.public.runs.Run,
        params: List[str] = [
            "mu_hidden",
            "sigma_hidden",
            "velocity_ratio_params/enforce",
            "velocity_ratio_params/target",
        ],
        base_path: Union[pathlib.Path, str] = pathlib.Path(
            "scdiffeq_data/wandb_downloads"
        ),
        n_target_files: int = 5,
    ) -> None:
        """ """
        self._run = run
        self._params = params
        self._n_target_files = n_target_files

        self.state = self._run.state
        self.name = self._run.name
        self.id = self._run.id
        if "seed" in self.config:
            self.seed = self.config["seed"]["value"]
        self._base_path = base_path

        self._get_params_of_interest()

    @property
    def save_name(self) -> str:
        if not hasattr(self, "_save_name"):
            name = self.name.replace("-", "_")
            self._save_name = f"{name}.{self.id}"
        return self._save_name

    @property
    def dir(self) -> pathlib.Path:
        if not hasattr(self, "_dir"):
            self._dir = self._base_path.joinpath(self.save_name)
            self._dir.mkdir(exist_ok=True, parents=True)
        return self._dir

    @property
    def config(self) -> Dict[str, Any]:
        if not hasattr(self, "_config"):
            self._config = json.loads(self._run.json_config)
        return self._config

    def _get_params_of_interest(self) -> None:
        for param in self._params:
            try:
                if not "/" in param:
                    setattr(self, param, self.config[param]["value"])
                else:
                    k, v = param.split("/")
                    setattr(self, param.replace("/", "_"), self.config[k]["value"][v])
            except Exception as e:
                logger.debug(
                    f"Could not set {param} to {self.config[param]['value']}, error: {e}"
                )

    def download_artifacts(self, force: bool = False) -> None:

        already_downloaded = 0
        newly_downloaded = 0

        self._DOWNLOAD_CALLED = True

        self._artifact_dirs = []

        for artifact in self._run.logged_artifacts():
            DOWNLOADED = False
            artifact_dir = self.dir.joinpath(artifact.name)
            artifact_dir = artifact_dir.parent.joinpath(
                artifact_dir.name.replace("-", "_").replace("__", "_").replace(":", ".")
            )
            if not artifact_dir.exists():
                artifact_dir.mkdir(exist_ok=True, parents=True)
            self._artifact_dirs.append(artifact_dir)
            n_files = len(list(artifact_dir.glob("*")))
            already_downloaded += n_files
            if n_files == artifact.file_count:
                DOWNLOADED = True
            if not DOWNLOADED or force:
                artifact.download(root=artifact_dir)
                newly_downloaded += artifact.file_count
        logger.info(
            f"{already_downloaded} files already downloaded. {newly_downloaded} files newly downloaded."
        )

    # def _get_history_path(self):
    #     history_dirs = [path for path in self._artifact_dirs if "history" in str(path.name)]

    #     if not history_dirs:
    #         raise FileNotFoundError(
    #             f"No artifact directory containing 'history' found for run {self.id} in {self.dir}. "
    #             f"Available artifact_dirs: {[p.name for p in self._artifact_dirs]}"
    #         )

    #     # Use the first directory found that contains "history"
    #     history_dir = history_dirs[0]
    #     if len(history_dirs) > 1:
    #         logger.warning(
    #             f"Multiple artifact directories containing 'history' found for run {self.id}: "
    #             f"{[p.name for p in history_dirs]}. Using '{history_dir.name}'."
    #         )

    #     parquet_files = sorted(list(history_dir.glob("*.parquet")))

    #     if not parquet_files:
    #         raise FileNotFoundError(
    #             f"No .parquet file found in history artifact directory '{history_dir}' for run {self.id}."
    #         )

    #     history_file_path = parquet_files[0]
    #     if len(parquet_files) > 1:
    #         logger.warning(
    #             f"Multiple .parquet files found in '{history_dir}': "
    #             f"{[p.name for p in parquet_files]}. Using '{history_file_path.name}'."
    #         )
        
    #     return history_file_path

    @property
    def history_path(self) -> pathlib.Path:
        return self.dir.joinpath("history.csv")

    @property
    def history(self) -> pd.DataFrame:
        if self._run.state == "finished" and (not hasattr(self, "_history")):
            if not self.history_path.exists():
                self._history = download_run_history(self._run, self.history_path)
            else:
                self._history = pd.read_csv(self.history_path)
        else:
            self._history = None
        return self._history


    def _select_paths(self, key) -> list:
        return [
            path.name.split(f"{key}_")[1]
            for path in self._artifact_dirs
            if key in str(path.name)
        ]

    @property
    def benchmarked_ckpts(self) -> List[str]:
        """Checkpoints for which there are saved ckpts and corresponding metrics"""
        if not hasattr(self, "_benchmarked_ckpts"):
            if not self._DOWNLOAD_CALLED:
                self.download_artifacts()
            self._ckpt_paths = self._select_paths(key="model_ckpt")
            self._metrics_paths = self._select_paths(key="fate_prediction_metrics")
            self._benchmarked_ckpts = list(
                set(self._metrics_paths) & set(self._ckpt_paths)
            )
        return self._benchmarked_ckpts

    @property
    def is_complete(self) -> bool:
        return len(self.benchmarked_ckpts) >= self._n_target_files

    def __repr__(self) -> str:
        return f"scDiffEq Run [wandb]"



import ABCParse
import cellplots
import matplotlib.cm as cm
import matplotlib.pyplot as plt

from ._box_plot import BoxPlot
from ._foreground_scatter import ForegroundScatter

from typing import List, Optional

class StyledBoxPlot(ABCParse.ABCParse):

    def __init__(
        self,
        colors: Optional[List[str]] = cm.tab20.colors,
        widths: Optional[float] = 0.85,
        scatter_kw={
            "alpha": 0.8,
            "s": 35,
        },
        suppress_scatter: bool = False,
        *args,
        **kwargs,
    ) -> None:

        self.__parse__(locals())

        self.BACKGROUND_KWARGS = {
            "showmeans": True,
            "showfliers": False,
            "meanline": True,
            "zorder": 1,
            "widths": self._widths,
        }
        self.FOREGROUND_KWARGS = {
            "showmeans": False,
            "showfliers": False,
            "meanline": False,
            "zorder": 2,
            "widths": self._widths,
        }

    @property
    def colors(self):
        if not hasattr(self, "_colors") or self._colors is None:
            self._colors = list(cm.tab20.colors)
        return self._colors

    def forward(self) -> None:
        bp = BoxPlot(boxplot_kwargs = self.FOREGROUND_KWARGS, colors=self._colors)
        self.fore_bp = bp(ax=self.ax, data = self._data, mode = "foreground")
        bp = BoxPlot(boxplot_kwargs = self.BACKGROUND_KWARGS, colors=self._colors)
        self.back_bp = bp(ax=self.ax, data = self._data, mode = "background")
        if not self._suppress_scatter:
            foreground_scatter = ForegroundScatter(
                colors = self._colors,
                scatter_kw = self._scatter_kw,
            )
            foreground_scatter(ax=self.ax, data = self._data)

        for en, (k, v) in enumerate(self._data.items()):
            if len(v) == 1:
                self.ax.scatter(
                    x = [int(en + 1)] * len(v),
                    y = v,
                    color = self._colors[en],
                    s = self._scatter_kw["s"],
                    alpha = self._scatter_kw["alpha"],
                )

    def __call__(self, data, ax: Optional[plt.Axes] = None) -> None:
        self._data = data
        if ax is None:
            fig, axes = cellplots.plot()
            ax = axes[0]
        self.ax = ax

        self.forward()


def boxplot(
    data: dict,
    ax: Optional[plt.Axes] = None,
    colors: Optional[List[str]] = cm.tab20.colors,
    widths: Optional[float] = 0.85,
    suppress_scatter: bool = False,
    scatter_kw={
        "alpha": 0.8,
        "s": 35,
    },
    *args,
    **kwargs,
) -> None:
    """
    Boxplot with foreground scatter points.
    """
    cls = StyledBoxPlot(
        colors = colors,
        widths = widths,
        scatter_kw = scatter_kw,
        suppress_scatter = suppress_scatter,
        *args,
        **kwargs,
    )
    cls(data, ax)
    return cls



# -- import packages: ---------------------------------------------------------
import pandas as pd

# -- import local dependencies: -----------------------------------------------
from ._conditions import conditions

# -- function: ----------------------------------------------------------------
def get_conditions(condition):
    defaults = {
        "mu_hidden": [512, 512],
        "sigma_hidden": [32, 32],
        "velocity_ratio": {"enforce": 100, "target": 2.5},
        "seed": -1,
    }
    updated_condition = {}
    for key, val in defaults.items():
        if key in condition:
            val = condition[key]
        updated_condition[key] = val
    E = updated_condition["velocity_ratio"]["enforce"]
    V = updated_condition["velocity_ratio"]["target"]
    updated_condition.pop("velocity_ratio")
    updated_condition["E"] = E
    updated_condition["V"] = V
    return updated_condition


def check_completion_tracker(tracked_completion_dict: dict):
    x = []
    for cond_i, done in tracked_completion_dict.items():
        for k, v in done.items():
            if not v is None:
                if not v in x:
                    x.append(v)
                else:
                    print(
                        f"Problem at: {cond_i}[{k}] - {v} has already been used elsewhere"
                    )


def track_completion(summary_df):
    CONDITIONS = conditions.copy()
    available_index = list(summary_df.index).copy()
    used_index = []
    extras = []  # we want to avoid these
    N_SEEDS = 5
    use_count = 0
    tracked_completion_dict = {}
    condition_list_set = []
    for i in range(N_SEEDS):
        for en, original_condition in CONDITIONS.items():
            if not en in tracked_completion_dict:
                tracked_completion_dict[en] = {}

            # Create a copy of the condition to avoid modifying the original in CONDITIONS
            condition = original_condition.copy()
            condition.update({"seed": i})
            updated_condition = get_conditions(condition)
            condition_list_set.append(updated_condition)

            # Initialize completion_tracker for this seed to None
            tracked_completion_dict[en][i] = None

            complete_bool = pd.concat(
                [
                    summary_df[key].astype(str) == str(val)
                    for key, val in updated_condition.items()
                ],
                axis=1,
            )
            corresponding_indices = complete_bool[complete_bool.all(1)].index.tolist()

            for ix in corresponding_indices:
                if not ix in used_index:
                    tracked_completion_dict[en][i] = ix
                    used_index.append(ix)
                    use_count += 1
                    break  # Use the first available index and stop searching
    check_completion_tracker(tracked_completion_dict=tracked_completion_dict)
    for av_ix in available_index:
        if not av_ix in used_index:
            extras.append(av_ix)
    n_extras = len(extras)
    print(
        f"Found {n_extras} extras: {extras} (Matched: {use_count} / {len(available_index)} available)"
    )

    extras_df = summary_df.loc[extras]

    return tracked_completion_dict, condition_list_set, extras_df





from ._fit_loss import fit_loss
from ._temporal_cmap import generate_temporal_cmap, temporal_colormap
from ._box_plot._styled_box_plot import boxplot

__all__ = [
    "fit_loss",
    "generate_temporal_cmap",
    "temporal_colormap",
    "boxplot",
]


# -- import packages: ---------------------------------------------------------
import ABCParse
import pandas as pd
import scdiffeq as sdq


# -- import local dependencies: ----------------------------------------------
from ._version_accuracy import VersionAccuracy


# -- set typing: --------------------------------------------------------------
from typing import List


# -- operational object class: ------------------------------------------------
class SummarizedCkpt(ABCParse.ABCParse):
    def __init__(self, project: sdq.io.Project):
        self.__parse__(locals())

    @property
    def versions(self) -> List[str]:
        return list(self._project._VERSION_PATHS.keys())

    def _get_version(self, version_name: str):
        return getattr(self._project, version_name)

    def __call__(self):

        BestResults = {}
        for version_name in self.versions:
            version = self._get_version(version_name=version_name)
            self.version_acc = VersionAccuracy(version=version)
            try:
                BestResults[version_name] = self.version_acc.best_test_from_train
            except:
                continue

        return pd.DataFrame(BestResults).T



from ._version_accuracy import VersionAccuracy
from ._summarized_ckpt import SummarizedCkpt
from ._best_ckpts import best_checkpoints
from ._summarized_loss import SummarizedLoss
from ._list_project_versions import list_project_versions
from ._screen_results import ScreenResults

__all__ = [
    "VersionAccuracy",
    "SummarizedCkpt",
    "best_checkpoints",
    "SummarizedLoss",
    "list_project_versions",
    "ScreenResults",
]




# -- import packages: ---------------------------------------------------------
import pandas as pd

# -- import local dependencies: -----------------------------------------------
from ._run import Run

class Runtime:
    def __init__(self, run: Run) -> None:
        self._run = run
        self._history = self._run.history[["epoch", "_timestamp", "_runtime"]].dropna()

    @property
    def total(self):
        """in seconds.
        includes setup (i.e., the time between zero and the
        first timestamp in history"""
        return self._run._run.summary["_runtime"]

    @property
    def training(self):
        """Just training, ignores data-related setup, etc."""
        t_init = self._history["_runtime"].iloc[0]
        t_final = self._history["_runtime"].iloc[-1]
        return t_final - t_init

    @property
    def per_epoch(self) -> pd.Series:
        """time for each epoch"""
        return self._history.groupby("epoch")["_runtime"].max().diff().dropna()

    @property
    def per_epoch_mean(self):
        """in seconds"""
        return self.per_epoch.mean()



# -- import packages: ---------------------------------------------------------
import adata_query
import anndata
import numpy as np
import scipy.sparse
import sklearn.decomposition
import sklearn.mixture
import sklearn.preprocessing

# -- import local dependencies: -----------------------------------------------
from ._map_time import map_time
from ._simulation_visualization import SimulationVisualization

# -- cls: ---------------------------------------------------------------------
class GaussianMixtureModelSampling:
    """
    Fit GMM to original data, then sample from the fitted distribution
    """

    def __init__(
        self,
        adata: anndata.AnnData,
        n_components: int = 50,
        n_gmm_components: int = 20,
        random_state: int = 42,
    ) -> None:

        self._adata = adata
        self._n_components = n_components
        self._n_gmm_components = n_gmm_components
        self._random_state = random_state

        self._preprocess()
        self._fit()
        self._visualization = SimulationVisualization(X_ref=self.X_pca)

    @property
    def n_cells(self) -> int:
        return self._adata.shape[0]

    @property
    def X_raw(self):
        if not hasattr(self, "_X_raw"):
            self._X_raw = adata_query.fetch(self._adata, key="X")
            if scipy.sparse.issparse(self._X_raw):
                if scipy.sparse.isspmatrix_csr(self._X_raw):
                    self._X_raw = self._X_raw.toarray()
                else:
                    self._X_raw = self._X_raw.A
        return self._X_raw

    def _preprocess(self) -> None:
        # -- scaler: ----------------------------------------------------
        self.scaler_model = sklearn.preprocessing.StandardScaler()
        self.X_scaled = self.scaler_model.fit_transform(self.X_raw)
        # -- pca: -------------------------------------------------------
        self.pca_model = sklearn.decomposition.PCA(n_components=self._n_components)
        self.X_pca = self.pca_model.fit_transform(self.X_scaled)

        # -- update reference adata: ------------------------------------
        self._adata.obsm["X_pca"] = self.X_pca

    def _fit(self) -> None:
        self.gmm = sklearn.mixture.GaussianMixture(
            n_components=self._n_gmm_components,
            random_state=self._random_state,
        )
        self.gmm.fit(self.X_pca)

    def _compute_distance(self, X_ref, X_sim) -> None:

        ref_mean, ref_std = X_ref.mean(0), X_ref.std(0)
        sim_mean, sim_std = X_sim.mean(0), X_sim.std(0)

        self._distance_mean = np.abs(ref_mean - sim_mean).max()
        self._distance_std = np.abs(ref_std - sim_std).max()

        print(f"mean ± std difference: {self._distance_mean:.3f}±{self._distance_std:.3f}")

    def __call__(
        self,
        target_size: int,
        report_validation: bool = True,
        plot: bool = False,
        n_neighbors: int = 20,
        time_key: str = "ct_pseudotime",
        constant: float = 1e-8,
    ) -> anndata.AnnData:
        X_sim, _ = self.gmm.sample(target_size)
        self.t_ref = self._adata.obs[time_key]
        self.t_sim = map_time(
            adata=self._adata,
            X_sim=X_sim,
            time_key=time_key,
            constant=constant,
            n_neighbors=n_neighbors,
        )

        if report_validation:
            self._compute_distance(X_ref=self.X_pca, X_sim=X_sim)
        if plot:
            self._visualization.plot(X_sim, t_ref=self.t_ref, t_sim=self.t_sim)

        adata_sim = anndata.AnnData(X_sim)
        adata_sim.obs[time_key] = self.t_sim

        return adata_sim



# -- import packages: ---------------------------------------------------------
import logging
import pathlib
import wandb

# -- import local dependencies: -----------------------------------------------
from ._run import Run

# -- set type hints: ----------------------------------------------------------
from typing import List, Optional, Union

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)

# -- cls: ---------------------------------------------------------------------
class WandbClient:

    def __init__(
        self,
        project: str,
        entity: str = "scDiffEq",
        base_path: Union[pathlib.Path, str] = pathlib.Path(
            "scdiffeq_data/wandb_downloads"
        ),
        params: List[str] = [
            "mu_hidden",
            "sigma_hidden",
            "velocity_ratio_params/enforce",
            "velocity_ratio_params/target",
        ],
        n_target_files: int = 5,
    ) -> None:

        self._project = project
        self._entity = entity
        self._all_runs = []
        self._complete_runs = []
        self._failed_runs = []
        self._base_path = pathlib.Path(base_path)
        self._params = params
        self._n_target_files = n_target_files
        self._init()

    def _init(self) -> None:

        self._api = wandb.Api()
        self._runs = self._api.runs(f"{self._entity}/{self._project}")

    def get_runs(self, refresh: bool = True, limit: Optional[int] = None) -> None:
        if refresh:
            self._init()
        self._all_runs = []
        self._complete_runs = []
        self._failed_runs = []
        if limit:
            iter_runs = self._runs[:limit]
        else:
            iter_runs = self._runs
        for run in iter_runs:
            try:
                sdq_run = Run(
                    run=run,
                    base_path=self._base_path,
                    params=self._params,
                    n_target_files=self._n_target_files,
                )
                sdq_run.download_artifacts()
                if sdq_run.is_complete:
                    self._complete_runs.append(sdq_run)
                self._all_runs.append(sdq_run)
            except Exception as e:
                logger.debug(
                    f"Failed to process run '{run.name}' (ID: {run.id}). Error: {e}",
                    exc_info=True
                )
                self._failed_runs.append(
                    {"id": run.id, "name": run.name, "error": str(e)}
                )

        n_all = len(self._all_runs)
        n_complete = len(self._complete_runs)
        n_failed = len(self._failed_runs)

        logger.info(f"Found {n_all} runs ({n_complete} complete runs, {n_failed} runs)")

    def _format_run_summary(self, min_benchmarked_ckpts: int = 5) -> list[str]:
        """Formats a summary of runs, including only those with a minimum number of benchmarked checkpoints.

        Args:
            min_benchmarked_ckpts: The minimum number of benchmarked checkpoints a run must have to be included.

        Returns:
            A list of strings, where each string is a formatted line of the run summary.
            The last string in the list is a total count of the runs included.
        """
        summary_lines = []
        count = 0
        for run in self._all_runs:
            if len(run.benchmarked_ckpts) >= min_benchmarked_ckpts:
                summary_lines.append(
                    f"{count:<3} | {run.name:<25} | [{len(run.benchmarked_ckpts):<2}] | {run.mu_hidden} {run.sigma_hidden}"
                )
                count += 1
        summary_lines.append(f"//total: {count}")
        return summary_lines

    def print_run_summary(self) -> None:
        print("\nFormatting run summary (default min_benchmarked_ckpts=5)...")
        summary_lines_default = self._format_run_summary()
        if summary_lines_default:
            print("\n--- Run Summary (Default) ---")
            for line in summary_lines_default:
                print(line)
        else:
            print("No runs met the default criteria for the summary.")

        print(f"\nTotal runs processed: {len(self._all_runs)}")
        print(f"Complete runs: {len(self._complete_runs)}")
        print(f"Failed runs: {len(self._failed_runs)}")

        if self._failed_runs:
            print("Details of failed runs:")
            for failed_run in self._failed_runs:
                print(
                    f"  ID: {failed_run['id']}, Name: {failed_run['name']}, Error: {failed_run['error']}"
                )



# -- import packages: ---------------------------------------------------------
import pandas as pd

# -- import local dependencies: -----------------------------------------------
from ._track_completion import track_completion

# -- set type hints: ----------------------------------------------------------
from typing import Dict, List, Tuple

# -- operational cls: ---------------------------------------------------------
class PlotDataAggr:
    """Hyperparameter scan plotting data"""

    def __init__(self, summary_df: pd.DataFrame) -> None:
        """"""
        self._summary_df = summary_df
        self._plot_data = {}
        self._fetch_all_plot_data()

    def print(self) -> None:
        for k, v in self._plot_data.items():
            print(f"{k:<15} {len(v)}")

    @property
    def tracked_completion(self) -> dict:
        if not hasattr(self, "_tracked_completion"):
            (
                self._tracked_completion,
                self._condition_list_set,
                self._extras,
            ) = track_completion(self._summary_df)
        return self._tracked_completion

    def _all_null(self, condition: Dict[int, int]) -> bool:
        return all([item == None for item in condition.values()])

    def _update(self, idx: int, condition: dict) -> None:
        if not self._all_null(condition):
            run_indices = [ix for ix in list(condition.values()) if not ix is None]
            train_result = self._summary_df.loc[run_indices]["train"].tolist()
            test_result = self._summary_df.loc[run_indices]["test"].tolist()
        else:
            train_result, test_result = [], []

        self._plot_data[f"cond_{idx}.train"] = train_result
        self._plot_data[f"cond_{idx}.test"] = test_result

    def _fetch_all_plot_data(self) -> None:
        for idx, condition in self.tracked_completion.items():
            self._update(idx=idx, condition=condition)

    def subset_plot_data_over_range(self, index_range: List[int]) -> Dict[str, List[float]]:
        plot_data_subset = {}
        for i in index_range:
            train = self._plot_data[f"cond_{i}.train"]
            test = self._plot_data[f"cond_{i}.test"]
            if len(train) == 0:
                train = [0]
            if len(test) == 0:
                test = [0]
            plot_data_subset.update({f"cond_{i}.train": train, f"cond_{i}.test": test})
        return plot_data_subset

    def __call__(
        self,
        indices: List[Tuple[int]] = [
            (0, 6),
            (6, 12),
            (12, 18),
            (18, 24),
            (24, 30),
            (30, 36),
            (36, 41),
        ],
    ):
        """ """

        plot_data = []
        for idx in indices:
            pl_subset = self.subset_plot_data_over_range(
                index_range=range(idx[0], idx[1])
            )
            plot_data.append(pl_subset)
        return plot_data

# -- function: ----------------------------------------------------------------
def aggregate_plot_data(
    summary_df: pd.DataFrame,
    indices: List[Tuple[int]] = [
        (0, 6),
        (6, 12),
        (12, 18),
        (18, 24),
        (24, 30),
        (30, 36),
        (36, 41),
    ],
) -> List[Dict[str, List[float]]]:
    aggr = PlotDataAggr(summary_df=summary_df)
    return aggr(indices=indices)



from ._aggregate_plot_data import aggregate_plot_data
from ._conditions import conditions
from ._track_completion import track_completion

__all__ = ["aggregate_plot_data", "conditions", "track_completion"]



# -- import packages: ---------------------------------------------------------
import cellplots
import numpy as np
import sklearn.decomposition


# -- cls: ---------------------------------------------------------------------
class SimulationVisualization:
    def __init__(self, X_ref: np.ndarray) -> None:
        """SimulationVisualization cls
        
        Args:
            X_ref (np.ndarray): Reference data
        """

        self._X_ref = X_ref

        # -- fit viz pca on reference cells: ---------------------
        self.viz_pca = sklearn.decomposition.PCA(n_components=2)
        self.ref_2d = self.viz_pca.fit_transform(self._X_ref)

    def plot(self, X_sim: np.ndarray, t_ref: np.ndarray, t_sim: np.ndarray) -> None:

        self.sim_2d = self.viz_pca.transform(X_sim)

        fig, axes = cellplots.plot(2, 2, wspace=0.1)

        pl_data = [self.ref_2d, self.sim_2d]
        color = [t_ref, t_sim]

        for en, ax in enumerate(axes):
            c = color[en]
            c_idx = np.argsort(c)
            c = c[c_idx]
            x, y = pl_data[en][c_idx, 0], pl_data[en][c_idx, 1]
            ax.scatter(x, y, c=c, cmap="plasma_r", ec="None", rasterized=True)



# -- import packages: ---------------------------------------------------------
import numpy as np

# -- cls: ---------------------------------------------------------------------
class ForegroundScatter:
    def __init__(self, colors, scatter_kw) -> None:
        """Foreground scatter points."""
        self.colors = colors
        self.scatter_kw = scatter_kw

    def forward(self, en, key, val) -> None:

        x = [key] * len(val)

        if len(x) > 1:
            x_vals = en + 1 + (np.random.random(len(x)) - 0.5) / 5
        else:
            x_vals = en + 1

        self.ax.scatter(
            x_vals,
            val,
            color=self.colors[en],
            zorder=0,
            ec="None",
            rasterized=False,
            **self.scatter_kw,
        )

    def __call__(self, ax, data) -> None:

        self.ax = ax

        for en, (key, val) in enumerate(data.items()):
            self.forward(en, key, val)



import pandas as pd

from typing import List

# -- import local dependencies: -----------------------------------------------
from ._run import Run
from ._summarize_run_accuracy import summarize_run_accuracy
from ._get_best_ckpt import get_best_ckpt

# -- function: ----------------------------------------------------------------
def summarize_runs(complete_runs: List[Run]) -> pd.DataFrame:
    best = []
    for run in complete_runs:
        acc_df = summarize_run_accuracy(run)
        best_ckpt = get_best_ckpt(acc_df)
        best_ckpt.update(
            {
                "mu_hidden": str(run.mu_hidden),
                "sigma_hidden": str(run.sigma_hidden),
                "E": str(run.velocity_ratio_params_enforce),
                "V": str(run.velocity_ratio_params_target),
                "seed": run.seed,
                "name": run.name,
            }
        )
        best.append(best_ckpt)
    df = pd.DataFrame(best).sort_values("test", ascending=False)
    df = df.rename({"epoch": "best_epoch"}, axis=1)
    df = df[
        [
            "name",
            "mu_hidden",
            "sigma_hidden",
            "V",
            "E",
            "seed",
            "best_epoch",
            "train",
            "test",
        ]
    ].copy()
    df = df.reset_index(drop=True).copy()
    return df



def list_project_versions(project):
    versions = []
    for item in project.__dir__():
        if item.startswith("version"):
            try:
                versions.append(getattr(project, item))
            except Exception as e:
                print(f"Problem with: {item}: {e}")
    return versions



from ._run import Run
from ._wandb_client import WandbClient
from ._get_best_ckpt import get_best_ckpt
from ._summarize_run_accuracy import summarize_run_accuracy
from ._summarize_runs import summarize_runs
from ._download_run_history import download_run_history

__all__ = [
    "Run",
    "WandbClient",
    "get_best_ckpt",
    "summarize_run_accuracy",
    "summarize_runs",
    "download_run_history",
]



# -- import packages: ---------------------------------------------------------
import anndata
import numpy as np
import scdiffeq as sdq


# -- cls: ---------------------------------------------------------------------
class TimeMapping:
    def __init__(
        self,
        adata: anndata.AnnData,
        n_neighbors: int = 20,
        time_key: str = "ct_pseudotime",
        constant: float = 1e-8,
    ) -> None:
        """TimeMapping cls
        
        Args:
            adata (anndata.AnnData): AnnData object
            n_neighbors (int): Number of neighbors
            time_key (str): Key for the time variable
            constant (float): Constant for the weights
        """
        self._adata = adata
        self._n_cells = adata.shape[0]
        self._n_neighbors = n_neighbors
        self._time_key = time_key
        self._constant = constant
        self._t_sim = []

    @property
    def t_ref(self):
        if not hasattr(self, "_t_ref"):
            self._t_ref = self._adata.obs[self._time_key]
        return self._t_ref

    @property
    def kNN(self):
        if not hasattr(self, "_knn"):
            self._knn = sdq.tl.kNN(
                self._adata, n_neighbors=self._n_neighbors, use_key="X_pca"
            )
        return self._knn

    def _compute_weights(self, nn_dist):
        weights = 1.0 / (nn_dist + self._constant)
        return weights / weights.sum()

    def forward(self, x_nn, x_dist):

        if x_dist.min() < 1e-8:
            return self.t_ref[x_nn[x_dist.argmin()]]
        else:
            weights = self._compute_weights(x_dist)
            return np.sum(weights * self.t_ref[x_nn].values)

    def __call__(self, X_sim):

        X_nn, X_dist = self.kNN.query(X_query=X_sim, include_distances=True)

        for i in range(len(X_sim)):
            self._t_sim.append(self.forward(x_nn=X_nn[i], x_dist=X_dist[i]))
        return np.array(self._t_sim)

# -- function: ----------------------------------------------------------------
def map_time(
    adata: anndata.AnnData,
    X_sim: np.ndarray,
    n_neighbors: int = 20,
    time_key: str = "ct_pseudotime",
    constant: float = 1e-8,
):
    """
    Args:
        adata (anndata.AnnData)
        X_sim (np.ndarray)
        n_neighbors (int) = 20,
        time_key (str) = "ct_pseudotime",
        constant (float) = 1e-8,

    Returns:
        t_sim (np.ndarray)
    """
    time_mapping = TimeMapping(
        adata=adata,
        n_neighbors=n_neighbors,
        time_key=time_key,
        constant=constant,
    )
    return time_mapping(X_sim)



# -- import packages: ---------------------------------------------------------
import pandas as pd

# -- import local dependencies: -----------------------------------------------
from ._run import Run

# -- cls: ---------------------------------------------------------------------
class RunAccuracy:
    def __init__(self): ...

    def _convert_ckpt_name_to_epoch(self, ckpt_name):
        if "last_run_id" in ckpt_name:
            return 2500
        elif "on_train_end_run_id" in ckpt_name:
            return 2499
        else:
            return int(ckpt_name.split("epoch_")[1].split(".")[0])

    def _get_accuracy_csv(self, run, ckpt):
        p = list(run.dir.glob(f"*{ckpt}/accuracy.csv"))
        _df = pd.read_csv(p[0], index_col=0)
        _df.columns = [self._convert_ckpt_name_to_epoch(ckpt)]
        return _df

    def forward(self, run: Run) -> pd.DataFrame:
        df = pd.concat(
            [
                self._get_accuracy_csv(run, ckpt_name)
                for ckpt_name in run.benchmarked_ckpts
            ],
            axis=1,
        )
        return df[sorted(df.columns.tolist())].copy()

    def __call__(self, run: Run) -> pd.DataFrame:
        return self.forward(run=run)

# -- function: ----------------------------------------------------------------
def summarize_run_accuracy(run: Run) -> pd.DataFrame:
    run_accuracy = RunAccuracy()
    return run_accuracy(run=run)



# -- import packages: ---------------------------------------------------------
import pandas as pd

# -- set type hints: ----------------------------------------------------------
from typing import Dict, Union

# -- function: ----------------------------------------------------------------
def get_best_ckpt(
    run_accuracy_df: pd.DataFrame,
    train_col: str = "unique_train.all_fates",
    test_col: str = "unique_test.all_fates",
) -> Dict[str, Union[float, str]]:
    df = run_accuracy_df.copy()
    df = (
        df.loc[[train_col, test_col]]
        .rename({train_col: "train", test_col: "test"}, axis=0)
        .T.reset_index()
        .rename({"index": "epoch"}, axis=1)
        .copy()
    )
    return (
        df.sort_values(["train", "epoch"], ascending=[False, False])
        .reset_index(drop=True)
        .loc[0]
        .to_dict()
    )



from .__version__ import __version__

from . import computational_complexity
from . import fate_prediction
from . import parsers
from . import _plotting as pl
from . import _tools as tl
from . import metrics
from . import hp_scan
from . import wandb

__all__ = [
    "computational_complexity",
    "fate_prediction",
    "hp_scan",
    "parsers",
    "metrics",
    "pl",
    "tl",
    "wandb",
    "__version__",
]